"""Test project's batches create command."""
# pylint: disable=wrong-import-order, import-error
import io
import operator
import sys
from uuid import uuid4

from click import echo
from click.testing import CliRunner

from gencove.client import APIClient, APIClientError  # noqa: I100
from gencove.command.projects.cli import create_project_batch
from gencove.models import ProjectBatches
from gencove.tests.decorators import assert_authorization
from gencove.tests.filters import filter_jwt, replace_gencove_url_vcr
from gencove.tests.projects.vcr.filters import (
    filter_project_batches_request,
    filter_project_batches_response,
)
from gencove.tests.upload.vcr.filters import filter_volatile_dates
from gencove.tests.utils import get_vcr_response

import pytest

from vcr import VCR


@pytest.fixture(scope="module")
def vcr_config():
    """VCR configuration."""
    return {
        "cassette_library_dir": "gencove/tests/projects/vcr",
        "filter_headers": [
            "Authorization",
            "Content-Length",
            "User-Agent",
        ],
        "filter_post_data_parameters": [
            ("email", "email@example.com"),
            ("password", "mock_password"),
        ],
        "match_on": ["method", "scheme", "port", "path", "query"],
        "path_transformer": VCR.ensure_suffix(".yaml"),
        "before_record_request": [
            replace_gencove_url_vcr,
            filter_project_batches_request,
        ],
        "before_record_response": [
            filter_jwt,
            filter_volatile_dates,
            filter_project_batches_response,
        ],
    }


@pytest.mark.default_cassette("jwt-create.yaml")
@pytest.mark.vcr
@assert_authorization
def test_create_project_batches__missing_batch_type(credentials, mocker):
    """Test batch creation failure when batch type is not sent."""
    runner = CliRunner()
    mocked_create_project_batch = mocker.patch.object(
        APIClient, "create_project_batch"
    )
    res = runner.invoke(
        create_project_batch,
        [
            str(uuid4()),
            *credentials,
            "--batch-name",
            "foo bar",
        ],
    )
    assert res.exit_code == 1
    mocked_create_project_batch.assert_not_called()
    assert "You must provide value for --batch-type" in res.output


@pytest.mark.default_cassette("jwt-create.yaml")
@pytest.mark.vcr
@assert_authorization
def test_create_project_batches__missing_batch_name(credentials, mocker):
    """Test batch creation failure when batch name is not sent."""
    runner = CliRunner()
    mocked_create_project_batch = mocker.patch.object(
        APIClient, "create_project_batch"
    )
    res = runner.invoke(
        create_project_batch,
        [
            str(uuid4()),
            *credentials,
            "--batch-type",
            "hd777k",
        ],
    )
    assert res.exit_code == 1
    mocked_create_project_batch.assert_not_called()
    assert "You must provide value for --batch-name" in res.output


@pytest.mark.default_cassette("jwt-create.yaml")
@pytest.mark.vcr
@assert_authorization
def test_create_project_batches__bad_project_id(credentials, mocker):
    """Test batch creation failure when non-uuid string is used as project
    id.
    """
    runner = CliRunner()
    mocked_create_project_batch = mocker.patch.object(
        APIClient,
        "create_project_batch",
    )
    res = runner.invoke(
        create_project_batch,
        [
            "1111111",
            *credentials,
            "--batch-type",
            "hd777k",
            "--batch-name",
            "foo bar",
        ],
    )
    assert res.exit_code == 1
    mocked_create_project_batch.assert_not_called()
    assert "Project ID is not valid" in res.output


@pytest.mark.default_cassette("jwt-create.yaml")
@pytest.mark.vcr
@assert_authorization
def test_create_project_batches__not_owned_project(credentials, mocker):
    """Test batch creation failure when project is not owned."""
    runner = CliRunner()
    mocked_create_project_batch = mocker.patch.object(
        APIClient,
        "create_project_batch",
        side_effect=APIClientError(message="", status_code=403),
    )

    res = runner.invoke(
        create_project_batch,
        [
            str(uuid4()),
            *credentials,
            "--batch-type",
            "hd777k",
            "--batch-name",
            "foo bar",
        ],
    )
    assert res.exit_code == 1
    mocked_create_project_batch.assert_called_once()
    assert "You do not have the sufficient permission" in res.output


@pytest.mark.default_cassette("jwt-create.yaml")
@pytest.mark.vcr
@assert_authorization
def test_create_project_batches__duplicate_client_ids(credentials, mocker):
    """Test batch creation failure when there are samples that share same
    client id.
    """
    runner = CliRunner()
    mocked_create_project_batch = mocker.patch.object(
        APIClient,
        "create_project_batch",
        side_effect=APIClientError(message="", status_code=400),
    )

    res = runner.invoke(
        create_project_batch,
        [
            str(uuid4()),
            *credentials,
            "--batch-type",
            "hd777k",
            "--batch-name",
            "foo bar",
        ],
    )
    assert res.exit_code == 0
    mocked_create_project_batch.assert_called_once()
    assert "There was an error creating project batches" in res.output


@pytest.mark.vcr
@assert_authorization
def test_create_project_batches__success__with_sample_ids(
    batch_name,
    batch_type,
    credentials,
    mocker,
    project_id_batches,
    recording,
    sample_id_batches,
    vcr,
):  # pylint: disable=too-many-arguments
    """Test batch creation success when when sample ids are explicitly sent."""
    runner = CliRunner()
    if not recording:
        # Mock create_project_batch only if using the cassettes, since we
        # mock the return value.
        create_project_batch_response = get_vcr_response(
            "/api/v2/project-batches/", vcr, operator.contains
        )
        mocked_create_project_batch = mocker.patch.object(
            APIClient,
            "create_project_batch",
            return_value=ProjectBatches(**create_project_batch_response),
        )
    res = runner.invoke(
        create_project_batch,
        [
            project_id_batches,
            *credentials,
            "--batch-type",
            batch_type,
            "--batch-name",
            batch_name,
            "--sample-ids",
            sample_id_batches,
        ],
    )
    assert res.exit_code == 0
    if not recording:
        mocked_create_project_batch.assert_called_once()

        output_line = io.BytesIO()
        sys.stdout = output_line
        for response in ProjectBatches(
            **create_project_batch_response
        ).results:
            echo(
                "\t".join(
                    [
                        str(response.id),
                        response.last_status.created.isoformat(),
                        response.last_status.status,
                        response.batch_type,
                        response.name,
                    ]
                )
            )
        assert output_line.getvalue() == res.output.encode()


@pytest.mark.vcr
@assert_authorization
def test_create_project_batches__success__without_sample_ids(
    batch_name,
    batch_type,
    credentials,
    mocker,
    project_id_batches,
    recording,
    vcr,
):  # pylint: disable=too-many-arguments
    """Test batch creation success when when sample ids are not explicitly
    sent.
    """
    runner = CliRunner()
    if not recording:
        # Mock create_project_batch only if using the cassettes, since we
        # mock the return value.
        create_project_batch_response = get_vcr_response(
            "/api/v2/project-batches/", vcr, operator.contains
        )
        mocked_create_project_batch = mocker.patch.object(
            APIClient,
            "create_project_batch",
            return_value=ProjectBatches(**create_project_batch_response),
        )
    res = runner.invoke(
        create_project_batch,
        [
            project_id_batches,
            *credentials,
            "--batch-type",
            batch_type,
            "--batch-name",
            batch_name,
        ],
    )
    assert res.exit_code == 0
    if not recording:
        mocked_create_project_batch.assert_called_once()

        output_line = io.BytesIO()
        sys.stdout = output_line
        for response in ProjectBatches(
            **create_project_batch_response
        ).results:
            echo(
                "\t".join(
                    [
                        str(response.id),
                        response.last_status.created.isoformat(),
                        response.last_status.status,
                        response.batch_type,
                        response.name,
                    ]
                )
            )
        assert output_line.getvalue() == res.output.encode()



"""Python library which enables you to use Gencoves' research backend."""
import click

from gencove import version
from gencove.command.basespace import basespace
from gencove.command.download import download
from gencove.command.projects import projects
from gencove.command.s3_imports import s3
from gencove.command.samples import samples
from gencove.command.upload import upload
from gencove.command.uploads import uploads
from gencove.command.webhook import webhooks


@click.group()
@click.version_option(version=version.version())
def cli():
    """Gencove's command line interface."""


cli.add_command(basespace)
cli.add_command(download)
cli.add_command(upload)
cli.add_command(uploads)
cli.add_command(projects)
cli.add_command(samples)
cli.add_command(s3)
cli.add_command(webhooks)


if __name__ == "__main__":
    cli()



"""Test samples set metadata command."""

import json
from uuid import uuid4

from click.testing import CliRunner

from gencove.client import APIClient, APIClientError
from gencove.command.samples.cli import set_metadata


def test_set_metadata__bad_sample_id(mocker):
    """Test set metadata failure when non-uuid string is used as
    sample id.
    """
    runner = CliRunner()

    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_set_metadata = mocker.patch.object(
        APIClient,
        "set_metadata",
    )

    res = runner.invoke(
        set_metadata,
        [
            "1111111",
            "--email",
            "foo@bar.com",
            "--password",
            "123",
            "--json",
            '{"somekey":"somevalue"}',
        ],
    )

    assert res.exit_code == 1
    mocked_login.assert_called_once()
    mocked_set_metadata.assert_not_called()
    assert "Sample ID is not valid" in res.output


def test_set_metadata__not_owned_sample(mocker):
    """Test set metadata failure when sample is not owned."""
    mocked_response = {"detail": "Not found."}

    runner = CliRunner()

    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_set_metadata = mocker.patch.object(
        APIClient,
        "set_metadata",
        return_value=mocked_response,
        side_effect=APIClientError(message="", status_code=403),
    )

    res = runner.invoke(
        set_metadata,
        [
            str(uuid4()),
            "--email",
            "foo@bar.com",
            "--password",
            "123",
            "--json",
            '{"somekey":"somevalue"}',
        ],
    )
    assert res.exit_code == 1
    mocked_login.assert_called_once()
    mocked_set_metadata.assert_called_once()
    assert "You do not have the sufficient permission" in res.output


def test_set_metadata__bad_json(mocker):
    """Test sample set metadata bad json failure."""
    sample_id = str(uuid4())

    runner = CliRunner()

    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_set_metadata = mocker.patch.object(
        APIClient,
        "set_metadata",
    )

    res = runner.invoke(
        set_metadata,
        [
            sample_id,
            "--email",
            "foo@bar.com",
            "--password",
            "123",
            "--json",
            '{"somekey": "somevalue"',
        ],
    )
    assert res.exit_code == 1
    mocked_login.assert_called_once()
    mocked_set_metadata.assert_not_called()
    assert "Metadata JSON is not valid" in res.output


def test_set_metadata__success(mocker):
    """Test sample set metadata success."""
    sample_id = str(uuid4())

    runner = CliRunner()

    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_set_metadata = mocker.patch.object(
        APIClient,
        "set_metadata",
    )
    payload = '{"somekey": "somevalue"}'
    res = runner.invoke(
        set_metadata,
        [
            sample_id,
            "--email",
            "foo@bar.com",
            "--password",
            "123",
            "--json",
            payload,
        ],
    )
    assert res.exit_code == 0
    mocked_login.assert_called_once()
    mocked_set_metadata.assert_called_once_with(
        sample_id, json.loads(payload)
    )
    assert "Assigned metadata to a sample {}".format(sample_id) in res.output



"""Download command constants."""
import re
from enum import Enum, unique
from typing import Optional, Tuple, Union
from uuid import UUID

from pydantic import BaseModel  # pylint: disable=no-name-in-module

from gencove.constants import (  # noqa: I100
    DownloadTemplateParts,
    Optionals,
)


@unique
class SampleStatuses(Enum):
    """SampleStatuses as enum"""

    SUCCEEDED = "succeeded"
    FAILED = "failed"


@unique
class SampleArchiveStatuses(Enum):
    """SampleArchiveStatuses enum"""

    AVAILABLE = "available"
    RESTORED = "restored"


ALLOWED_STATUSES_RE = re.compile(
    "{}|{}".format(
        SampleStatuses.SUCCEEDED.value, SampleStatuses.FAILED.value
    ),
    re.IGNORECASE,
)
ALLOWED_ARCHIVE_STATUSES_RE = re.compile(
    "{}|{}".format(
        SampleArchiveStatuses.AVAILABLE.value,
        SampleArchiveStatuses.RESTORED.value,
    ),
    re.IGNORECASE,
)
KILOBYTE = 1024
MEGABYTE = 1024 * KILOBYTE
NUM_MB_IN_CHUNK = 3
CHUNK_SIZE = NUM_MB_IN_CHUNK * MEGABYTE


# pylint: disable=too-few-public-methods
class DownloadFilters(BaseModel):
    """DownloadFilters options"""

    # project_id and sample_ids should be UUID instead of Union[UUID, str]
    # some tests break on UUID because they are using arbitrary strings
    # fix tests and then modify
    project_id: Optional[Union[UUID, str]]
    sample_ids: Optional[Tuple[Union[UUID, str], ...]]
    file_types: Optional[Tuple[str, ...]]


# pylint: disable=too-few-public-methods
class DownloadOptions(Optionals):
    """DownloadOptions model"""

    skip_existing: Optional[bool]
    download_template: Optional[str]


DEFAULT_FILENAME_TOKEN = "{{{}}}".format(
    DownloadTemplateParts.DEFAULT_FILENAME.value
)


# pylint: disable=too-few-public-methods
class FilePrefix(BaseModel):
    """FilePrefix model"""

    dirs: str
    filename: str
    file_extension: str
    use_default_filename: bool


FILENAME_RE = re.compile("filename=(.+)")

# Leaving the line below since it is a great explaination of FILE_TYPES_MAPPER
# usage:
# FILE_TYPES_MAPPER = {"fastq-r1": "fastq-r1_R1", "fastq-r2": "fastq-r2_R2"}
FILE_TYPES_MAPPER = {}

QC_FILE_TYPE = "qc"
METADATA_FILE_TYPE = "metadata"



"""Merge project's VCF files executor."""
import backoff

from .utils import get_line
from ...base import Command
from ...utils import is_valid_uuid
from .... import client
from ....exceptions import ValidationError


class StatusMergedVCF(Command):
    """Executor for retrieving the status of a job for merging project's VCF
    files.
    """

    def __init__(
        self,
        project_id,
        credentials,
        options,
    ):
        super().__init__(credentials, options)
        self.project_id = project_id

    def initialize(self):
        """Initialize subcommand."""
        self.login()

    def validate(self):
        """Validate command input.

        Raises:
            ValidationError - if something is wrong with command parameters.
        """
        if is_valid_uuid(self.project_id) is False:
            raise ValidationError("Project ID is not valid. Exiting.")

    @backoff.on_exception(
        backoff.expo,
        (client.APIClientTimeout),
        max_tries=5,
        max_time=30,
    )
    def execute(self):
        """Get a status of a request to merge VCF files for a given project."""
        self.echo_debug(
            "Retrieving the status of a merge VCF files operation for "
            "project {}".format(self.project_id)
        )
        try:
            merge_details_status = self.api_client.retrieve_merged_vcf(
                self.project_id
            )
            self.echo_debug(merge_details_status)
            if merge_details_status.last_status.status == "failed":
                self.echo_warning("The job failed merging.")
            self.echo_data(get_line(merge_details_status))
        except client.APIClientError as err:
            self.echo_debug(err)
            if err.status_code == 400:
                self.echo_error(
                    "There was an error checking the status of "
                    "a merged VCF."
                )
            elif err.status_code == 404:
                self.echo_error(
                    "Project {} does not exist or there are no "
                    "running jobs associated with it.".format(self.project_id)
                )
            raise



"""VCR filters for download tests."""
from gencove.tests.filters import _replace_uuid_from_url


def filter_project_samples_request(request):
    """Filter project samples sensitive data from request."""
    return _replace_uuid_from_url(request, "project-samples")


def filter_sample_quality_controls_request(request):
    """Filter sample-quality-controls sensitive data from request."""
    return _replace_uuid_from_url(request, "sample-quality-controls")


def filter_sample_metadata_request(request):
    """Filter sample-metadata sensitive data from request."""
    return _replace_uuid_from_url(request, "sample-metadata")



"""Test project's batch types list command."""
# pylint: disable=wrong-import-order, import-error
import io
import operator
import sys
from uuid import uuid4

from click import echo
from click.testing import CliRunner

from gencove.client import APIClient, APIClientTimeout  # noqa: I100
from gencove.command.projects.cli import list_project_batch_types
from gencove.models import ProjectBatchTypes
from gencove.tests.decorators import assert_authorization
from gencove.tests.filters import filter_jwt, replace_gencove_url_vcr
from gencove.tests.projects.vcr.filters import (
    filter_get_project_batch_types_request,
)
from gencove.tests.upload.vcr.filters import filter_volatile_dates
from gencove.tests.utils import get_vcr_response

import pytest

from vcr import VCR


@pytest.fixture(scope="module")
def vcr_config():
    """VCR configuration."""
    return {
        "cassette_library_dir": "gencove/tests/projects/vcr",
        "filter_headers": [
            "Authorization",
            "Content-Length",
            "User-Agent",
        ],
        "filter_post_data_parameters": [
            ("email", "email@example.com"),
            ("password", "mock_password"),
        ],
        "match_on": ["method", "scheme", "port", "path", "query"],
        "path_transformer": VCR.ensure_suffix(".yaml"),
        "before_record_request": [
            replace_gencove_url_vcr,
            filter_get_project_batch_types_request,
        ],
        "before_record_response": [
            filter_jwt,
            filter_volatile_dates,
        ],
    }


@pytest.mark.vcr
@assert_authorization
def test_list_project_batch_types__empty(
    credentials, mocker, project_id, recording, vcr
):  # pylint: disable=unused-argument
    """Test project has not batch types."""
    runner = CliRunner()
    if not recording:
        # Mock get_project_batch_types only if using the cassettes, since we
        # mock the return value.
        get_project_batch_types_response = get_vcr_response(
            "/api/v2/project-batch-types/", vcr, operator.contains
        )
        mocked_get_project_batch_types = mocker.patch.object(
            APIClient,
            "get_project_batch_types",
            return_value=ProjectBatchTypes(
                **get_project_batch_types_response
            ),
        )

    res = runner.invoke(
        list_project_batch_types,
        [project_id, *credentials],
    )
    assert res.exit_code == 0
    if not recording:
        mocked_get_project_batch_types.assert_called_once()
    assert res.output == ""


@pytest.mark.default_cassette("jwt-create.yaml")
@pytest.mark.vcr
@assert_authorization
def test_list_project_batch_types__not_empty_slow_response_retry(
    mocker, credentials
):
    """Test project batch types being outputed to the shell."""
    runner = CliRunner()
    mocked_get_project_batch_types = mocker.patch.object(
        APIClient,
        "get_project_batch_types",
        side_effect=APIClientTimeout("Could not connect to the api server"),
    )

    res = runner.invoke(
        list_project_batch_types,
        [str(uuid4()), *credentials],
    )
    assert res.exit_code == 1
    assert mocked_get_project_batch_types.call_count == 2


@pytest.mark.vcr
@assert_authorization
def test_list_project_batch_types__not_empty(
    mocker, credentials, project_id_batches, recording, vcr
):
    """Test project batch types being outputed to the shell."""
    runner = CliRunner()
    if not recording:
        # Mock get_project_batch_types only if using the cassettes, since we
        # mock the return value.
        get_project_batch_types_response = get_vcr_response(
            "/api/v2/project-batch-types/", vcr, operator.contains
        )
        mocked_get_project_batch_types = mocker.patch.object(
            APIClient,
            "get_project_batch_types",
            return_value=ProjectBatchTypes(
                **get_project_batch_types_response
            ),
        )

    res = runner.invoke(
        list_project_batch_types,
        [project_id_batches, *credentials],
    )
    assert res.exit_code == 0
    assert res.output != ""
    if not recording:
        mocked_get_project_batch_types.assert_called_once()

        output_line = io.BytesIO()
        sys.stdout = output_line
        batches_output = [
            f"{batch_type['key']}\t{batch_type['description']}"
            for batch_type in get_project_batch_types_response["results"]
        ]
        echo("\n".join(batches_output))
        assert output_line.getvalue() == res.output.encode()



"""Constants for s3-import command."""
from typing import Optional

from gencove.constants import Optionals


class S3ImportOptionals(Optionals):
    """S3ImportOptionals model"""

    metadata_json: Optional[str]



"""Download sample file subcommand."""

import backoff

import requests

from gencove.command.download.constants import (  # noqa: I100
    ALLOWED_ARCHIVE_STATUSES_RE,
)

from .utils import download_file, fatal_process_sample_error
from ...base import Command
from ...utils import is_valid_uuid
from .... import client
from ....exceptions import ValidationError


class DownloadFile(Command):
    """Download file command executor."""

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        sample_id,
        file_type,
        destination,
        credentials,
        options,
        no_progress,
    ):
        super().__init__(credentials, options)
        self.sample_id = sample_id
        self.file_type = file_type
        self.destination = destination
        self.no_progress = no_progress

    def initialize(self):
        """Initialize download-file subcommand."""
        self.login()

    def validate(self):
        """Validate command input.

        Raises:
            ValidationError - if something is wrong with command parameters.
        """

        if is_valid_uuid(self.sample_id) is False:
            raise ValidationError("Sample ID is not valid. Exiting.")

    def execute(self):
        self.echo_debug("Retrieving sample file:")

        try:
            self.get_file()
        except client.APIClientError as err:
            self.echo_debug(err)
            if err.status_code == 404:
                self.echo_error(
                    "Sample {} file {} does not exist or you do not have "
                    "permission required to access it.".format(
                        self.sample_id, self.file_type
                    )
                )
            raise

    @backoff.on_exception(
        backoff.expo,
        requests.exceptions.HTTPError,
        giveup=fatal_process_sample_error,
        max_tries=10,
    )
    def get_file(self):
        """Retrieves file and writes it to destination"""
        try:
            sample = self.api_client.get_sample_details(self.sample_id)
        except client.APIClientError:
            self.echo_warning(
                "Sample with id {} not found. "
                "Are you using client id instead of sample id?".format(
                    self.sample_id
                )
            )
            return

        self.echo_debug(
            "Processing sample id {}, status {}".format(
                sample.id, sample.last_status.status
            )
        )

        if not ALLOWED_ARCHIVE_STATUSES_RE.match(
            sample.archive_last_status.status
        ):
            raise ValidationError(
                "Sample with id {} is archived and cannot be downloaded - "
                "please restore the sample and try again.".format(sample.id)
            )

        file_to_download = None

        for sample_file in sample.files:
            # pylint: disable=E0012,C0330
            if self.file_type == sample_file.file_type:
                file_to_download = sample_file
                break

        if file_to_download is None or file_to_download.download_url is None:
            self.echo_warning(
                "File not found for sample with id {} and file type {}".format(
                    self.sample_id, self.file_type
                )
            )
        else:
            download_file(
                self.destination,
                file_to_download.download_url,
                self.no_progress,
            )



"""Test project's restore samples command."""
# pylint: disable=wrong-import-order, import-error
import operator
from uuid import uuid4

from click.testing import CliRunner

from gencove.client import (
    APIClient,
    APIClientError,
)  # noqa: I100
from gencove.command.projects.cli import restore_project_samples
from gencove.tests.decorators import assert_authorization
from gencove.tests.filters import filter_jwt, replace_gencove_url_vcr
from gencove.tests.projects.vcr.filters import (
    filter_post_project_restore_samples,
)
from gencove.tests.upload.vcr.filters import filter_volatile_dates
from gencove.tests.utils import get_vcr_response

import pytest

from vcr import VCR


@pytest.fixture(scope="module")
def vcr_config():
    """VCR configuration."""
    return {
        "cassette_library_dir": "gencove/tests/projects/vcr",
        "filter_headers": [
            "Authorization",
            "Content-Length",
            "User-Agent",
        ],
        "filter_post_data_parameters": [
            ("email", "email@example.com"),
            ("password", "mock_password"),
        ],
        "match_on": ["method", "scheme", "port", "path", "query"],
        "path_transformer": VCR.ensure_suffix(".yaml"),
        "before_record_request": [
            replace_gencove_url_vcr,
            filter_post_project_restore_samples,
        ],
        "before_record_response": [
            filter_jwt,
            filter_volatile_dates,
        ],
    }


@pytest.mark.default_cassette("jwt-create.yaml")
@pytest.mark.vcr
@assert_authorization
def test_restore_project_samples__bad_project_id(mocker, credentials):
    """Test restore project samples when non-uuid string is used as project
    id."""
    runner = CliRunner()
    mocked_restore_project_samples = mocker.patch.object(
        APIClient,
        "restore_project_samples",
    )
    res = runner.invoke(
        restore_project_samples,
        [
            "1111111",
            *credentials,
            "--sample-ids",
            "11111111-1111-1111-1111-111111111111,22222222-2222-2222-2222-222222222222",  # noqa
        ],
    )
    assert res.exit_code == 1
    mocked_restore_project_samples.assert_not_called()
    assert "Project ID is not valid" in res.output


@pytest.mark.default_cassette("jwt-create.yaml")
@pytest.mark.vcr
@assert_authorization
def test_restore_project_samples__not_owned_project(credentials, mocker):
    """Test restore project samples failure when project is not owned."""
    runner = CliRunner()
    mocked_restore_project_samples = mocker.patch.object(
        APIClient,
        "restore_project_samples",
        side_effect=APIClientError(message="", status_code=403),
    )

    res = runner.invoke(
        restore_project_samples,
        [
            str(uuid4()),
            *credentials,
            "--sample-ids",
            "11111111-1111-1111-1111-111111111111,22222222-2222-2222-2222-222222222222",  # noqa
        ],
    )
    assert res.exit_code == 1
    mocked_restore_project_samples.assert_called_once()
    assert "You do not have the sufficient permission" in res.output


@pytest.mark.vcr
@assert_authorization
def test_restore_project_samples__success__empty_sample_ids(
    credentials, mocker, project_id, recording, vcr
):
    """Test restore project samples success when an empty list of sample ids
    is sent."""
    runner = CliRunner()
    if not recording:
        # Mock restore_project_samples only if using the cassettes, since we
        # mock the return value.
        restore_project_samples_response = get_vcr_response(
            "/api/v2/project-restore-samples/", vcr, operator.contains
        )
        mocked_restore_project_samples = mocker.patch.object(
            APIClient,
            "restore_project_samples",
            return_value=restore_project_samples_response,
        )

    res = runner.invoke(
        restore_project_samples,
        [
            project_id,
            *credentials,
            "--sample-ids",
            "",
        ],
    )
    assert res.exit_code == 0
    if not recording:
        mocked_restore_project_samples.assert_called_once()


@pytest.mark.default_cassette("jwt-create.yaml")
@pytest.mark.vcr
@assert_authorization
def test_restore_project_samples__invalid_sample_ids(credentials, mocker):
    """Test restore project samples failure when an empty list of sample ids
    is sent."""

    runner = CliRunner()
    mocked_restore_project_samples = mocker.patch.object(
        APIClient, "restore_project_samples"
    )

    res = runner.invoke(
        restore_project_samples,
        [
            str(uuid4()),
            *credentials,
            "--sample-ids",
            "1111,222",
        ],
    )
    assert res.exit_code == 1
    mocked_restore_project_samples.assert_not_called()
    assert "Not all sample IDs are valid" in res.output


@pytest.mark.vcr
@assert_authorization
def test_restore_project_samples__sample_not_in_project(
    credentials, mocker, project_id, recording, vcr
):
    """Test restore project samples with sample not in project."""

    runner = CliRunner()
    if not recording:
        # Mock restore_project_samples only if using the cassettes, since we
        # mock the return value.
        restore_project_samples_response = get_vcr_response(
            "/api/v2/project-restore-samples/",
            vcr,
            operator.contains,
            just_body=False,
        )
        mocked_restore_project_samples = mocker.patch.object(
            APIClient,
            "restore_project_samples",
            side_effect=APIClientError(
                message=restore_project_samples_response["body"]["string"],
                status_code=restore_project_samples_response["status"][
                    "code"
                ],
            ),
        )
    res = runner.invoke(
        restore_project_samples,
        [
            project_id,
            *credentials,
            "--sample-ids",
            "11111111-1111-1111-1111-111111111111",
        ],
    )
    assert res.exit_code == 0
    if not recording:
        mocked_restore_project_samples.assert_called_once()
    assert (
        "There was an error requesting project samples restore" in res.output
    )


@pytest.mark.vcr
@assert_authorization
def test_restore_project_samples__success(
    archived_sample, credentials, mocker, project_id, recording, vcr
):  # pylint: disable=too-many-arguments
    """Test restore project samples success."""
    runner = CliRunner()
    if not recording:
        # Mock restore_project_samples only if using the cassettes, since we
        # mock the return value.
        restore_project_samples_response = get_vcr_response(
            "/api/v2/project-restore-samples/", vcr, operator.contains
        )
        mocked_restore_project_samples = mocker.patch.object(
            APIClient,
            "restore_project_samples",
            return_value=restore_project_samples_response,
        )

    res = runner.invoke(
        restore_project_samples,
        [
            project_id,
            *credentials,
            "--sample-ids",
            archived_sample,
        ],
    )
    assert res.exit_code == 0
    if not recording:
        mocked_restore_project_samples.assert_called_once()
    assert "Request to restore samples accepted" in res.output



"""Import samples from S3 to a project subcommand."""

import json

from ...base import Command
from ...utils import is_valid_json, is_valid_uuid
from .... import client
from ....exceptions import ValidationError


class S3Import(Command):
    """S3 import command executor."""

    def __init__(self, s3_uri, project_id, credentials, options):
        super().__init__(credentials, options)
        self.s3_uri = s3_uri
        self.project_id = project_id
        self.metadata_json = options.metadata_json

    def initialize(self):
        """Initialize s3-import subcommand."""
        self.login()

    def validate(self):
        """Validate command input.

        Raises:
            ValidationError - if something is wrong with command parameters.
        """
        if is_valid_uuid(self.project_id) is False:
            raise ValidationError("Project ID is not valid. Exiting.")
        if self.metadata_json and is_valid_json(self.metadata_json) is False:
            raise ValidationError("Metadata JSON is not valid. Exiting.")

    # no retry for timeouts in order to avoid duplicate heavy operations on
    # the backend
    def execute(self):
        """Make a request to import samples from S3 to a given
        project.
        """
        self.echo_debug(
            "Import samples from {} to project {}".format(
                self.s3_uri, self.project_id
            )
        )

        try:
            # prepare metadata
            metadata = None
            if self.metadata_json is not None:
                metadata = json.loads(self.metadata_json)
                self.echo_info("Assigning metadata to the imported samples.")
            s3_uri_import = self.api_client.s3_uri_import(
                s3_uri=self.s3_uri,
                project_id=self.project_id,
                metadata=metadata,
            )
            self.echo_debug(s3_uri_import)
            self.echo_info("Request to import samples from S3 accepted.")
        except client.APIClientError:
            self.echo_error("There was an error importing from S3.")
            raise



"""Merge project's VCF files executor."""
from ..status_merged_vcf.utils import get_line
from ...base import Command
from ...utils import is_valid_uuid
from .... import client
from ....exceptions import ValidationError


class CreateMergedVCF(Command):
    """Merge project's VCF files executor."""

    def __init__(
        self,
        project_id,
        credentials,
        options,
    ):
        super().__init__(credentials, options)
        self.project_id = project_id

    def initialize(self):
        """Initialize subcommand."""
        self.login()

    def validate(self):
        """Validate command input.

        Raises:
            ValidationError - if something is wrong with command parameters.
        """
        if is_valid_uuid(self.project_id) is False:
            raise ValidationError("Project ID is not valid. Exiting.")

    # no retry for timeouts in order to avoid duplicate heavy operations on
    # the backend
    def execute(self):
        """Make a request to merge VCF files for a given project."""
        self.echo_debug(
            "Merging VCF files for project {}".format(self.project_id)
        )
        try:
            created_merge_details = self.api_client.create_merged_vcf(
                self.project_id
            )
            self.echo_debug(created_merge_details)
            self.echo_info(
                "Issued merge request for project {}".format(self.project_id),
            )
            self.echo_data(get_line(created_merge_details))
        except client.APIClientError as err:
            self.echo_debug(err)
            if err.status_code == 400:
                self.echo_error("There was an error creating merged VCF.")
            elif err.status_code == 404:
                self.echo_error(
                    "Project {} does not exist.".format(self.project_id)
                )
            raise



"""Set sample metadata subcommand."""

import json

from ...base import Command
from ...utils import is_valid_json, is_valid_uuid
from .... import client
from ....exceptions import ValidationError


class SetMetadata(Command):
    """Set metadata command executor."""

    def __init__(self, sample_id, json_metadata, credentials, options):
        super().__init__(credentials, options)
        self.sample_id = sample_id
        self.json_metadata = json_metadata

    def initialize(self):
        """Initialize set-metadata subcommand."""
        self.login()

    def validate(self):
        """Validate command input.

        Raises:
            ValidationError - if something is wrong with command parameters.
        """

        if is_valid_uuid(self.sample_id) is False:
            raise ValidationError("Sample ID is not valid. Exiting.")

        if is_valid_json(self.json_metadata) is False:
            raise ValidationError("Metadata JSON is not valid. Exiting.")

    # no retry for timeouts in order to avoid duplicate heavy operations on
    # the backend
    def execute(self):
        """Make a request to assign given metadata to a specified sample."""

        self.echo_debug(
            "Assigning metadata to a sample {}".format(self.sample_id)
        )

        try:
            metadata_api = None
            if self.json_metadata is not None:
                metadata_api = json.loads(self.json_metadata)
            assigned_metadata = self.api_client.set_metadata(
                self.sample_id, metadata_api
            )
            self.echo_debug(assigned_metadata)
            self.echo_info(
                "Assigned metadata to a sample {}".format(self.sample_id)
            )
        except client.APIClientError as err:
            self.echo_debug(err)
            if err.status_code == 400:
                self.echo_error("There was an error assigning metadata.")
            elif err.status_code == 404:
                self.echo_error(
                    "Sample {} does not exist.".format(self.sample_id)
                )
            raise



"""Sample get metadata shell command definition."""
import click

from gencove.command.common_cli_options import add_options, common_options
from gencove.constants import Credentials, Optionals

from .main import GetMetadata


@click.command("get-metadata")
@click.argument("sample_id")
@click.option(
    "--output-filename",
    help="Output filename for the metadata.",
    type=click.Path(),
    default="-",
    required=False,
)
@add_options(common_options)
# pylint: disable=too-many-arguments
def get_metadata(
    sample_id,
    output_filename,
    host,
    email,
    password,
    api_key,
):
    """Get sample metadata."""
    GetMetadata(
        sample_id,
        output_filename,
        Credentials(email=email, password=password, api_key=api_key),
        Optionals(host=host),
    ).run()



"""Project get merged VCF shell command definition."""
import click

from gencove.command.common_cli_options import add_options, common_options
from gencove.constants import Credentials, Optionals

from .main import GetMergedVCF


@click.command("get-merged-vcf")
@click.argument("project_id")
@click.option(
    "--output-filename",
    help="Output filename for merged VCF file.",
    type=click.Path(),
    required=False,
    default=None,
)
@add_options(common_options)
@click.option(
    "--no-progress",
    is_flag=True,
    help="If specified, no progress bar is shown.",
)
# pylint: disable=too-many-arguments
def get_merged_vcf(
    project_id,
    output_filename,
    host,
    email,
    password,
    api_key,
    no_progress,
):
    """Download merged VCF file in a project."""
    GetMergedVCF(
        project_id,
        output_filename,
        Credentials(email=email, password=password, api_key=api_key),
        Optionals(host=host),
        no_progress,
    ).run()



"""Test create merged project's VCF file command."""

import io
import sys
from uuid import uuid4

from click import echo
from click.testing import CliRunner

from gencove.client import APIClient, APIClientError  # noqa: I100
from gencove.command.projects.cli import create_merged_vcf
from gencove.models import ProjectMergeVCFs


def test_create_merged_vcf__bad_project_id(mocker):
    """Test create merged file failure when non-uuid string is used as
    project id.
    """
    runner = CliRunner()

    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_create_merged_vcf = mocker.patch.object(
        APIClient,
        "create_merged_vcf",
        return_value=ProjectMergeVCFs(id=uuid4()),
    )

    res = runner.invoke(
        create_merged_vcf,
        [
            "1111111",
            "--email",
            "foo@bar.com",
            "--password",
            "123",
        ],
    )

    assert res.exit_code == 1
    mocked_login.assert_called_once()
    mocked_create_merged_vcf.assert_not_called()
    assert "Project ID is not valid" in res.output


def test_create_merged_vcf__not_owned_project(mocker):
    """Test create merged file failure when project is not owned."""
    mocked_response = {"detail": "Not found."}

    runner = CliRunner()

    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_create_merged_vcf = mocker.patch.object(
        APIClient,
        "create_merged_vcf",
        return_value=mocked_response,
        side_effect=APIClientError(message="", status_code=403),
    )

    res = runner.invoke(
        create_merged_vcf,
        [
            str(uuid4()),
            "--email",
            "foo@bar.com",
            "--password",
            "123",
        ],
    )

    assert res.exit_code == 1
    mocked_login.assert_called_once()
    mocked_create_merged_vcf.assert_called_once()
    assert "You do not have the sufficient permission" in res.output


def test_create_merged_vcf__already_running(mocker):
    """Test create merged file failure when the job is alredy running."""
    project_id = str(uuid4())
    message = "Merge VCFs for project {} is already running".format(
        project_id
    )
    mocked_response = [message]

    runner = CliRunner()

    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_create_merged_vcf = mocker.patch.object(
        APIClient,
        "create_merged_vcf",
        return_value=mocked_response,
        side_effect=APIClientError(message=message, status_code=400),
    )

    res = runner.invoke(
        create_merged_vcf,
        [
            project_id,
            "--email",
            "foo@bar.com",
            "--password",
            "123",
        ],
    )

    assert res.exit_code == 1
    mocked_login.assert_called_once()
    mocked_create_merged_vcf.assert_called_once()
    assert "is already running" in res.output


def test_create_merged_vcf__not_able(mocker):
    """Test create merged file failure when it cannot be done."""
    project_id = str(uuid4())
    mocked_response = [
        "You attempted to merge VCFs for a project that cannot do that"
    ]

    runner = CliRunner()

    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_create_merged_vcf = mocker.patch.object(
        APIClient,
        "create_merged_vcf",
        return_value=mocked_response,
        side_effect=APIClientError(
            message=mocked_response[0], status_code=400
        ),
    )

    res = runner.invoke(
        create_merged_vcf,
        [
            project_id,
            "--email",
            "foo@bar.com",
            "--password",
            "123",
        ],
    )

    assert res.exit_code == 1
    mocked_login.assert_called_once()
    mocked_create_merged_vcf.assert_called_once()
    assert (
        "You attempted to merge VCFs for a project that cannot do that"
        in res.output
    )


def test_create_merged_vcf__no_samples(mocker):
    """Test create merged file failure when there are no samples."""
    project_id = str(uuid4())
    mocked_response = ["No samples found in project"]

    runner = CliRunner()

    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_create_merged_vcf = mocker.patch.object(
        APIClient,
        "create_merged_vcf",
        return_value=mocked_response,
        side_effect=APIClientError(
            message=mocked_response[0], status_code=400
        ),
    )

    res = runner.invoke(
        create_merged_vcf,
        [
            project_id,
            "--email",
            "foo@bar.com",
            "--password",
            "123",
        ],
    )

    assert res.exit_code == 1
    mocked_login.assert_called_once()
    mocked_create_merged_vcf.assert_called_once()
    assert "No samples found in project" in res.output


def test_create_merged_vcf__up_to_date(mocker):
    """Test create merged file failure when attempting to create another one
    that is the same.
    """

    project_id = str(uuid4())
    message = "Merged VCF file for project {} is up-to-date".format(
        project_id
    )
    mocked_response = [message]

    runner = CliRunner()

    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_create_merged_vcf = mocker.patch.object(
        APIClient,
        "create_merged_vcf",
        return_value=mocked_response,
        side_effect=APIClientError(message=message, status_code=400),
    )

    res = runner.invoke(
        create_merged_vcf,
        [
            project_id,
            "--email",
            "foo@bar.com",
            "--password",
            "123",
        ],
    )

    assert res.exit_code == 1
    mocked_login.assert_called_once()
    mocked_create_merged_vcf.assert_called_once()
    assert "is up-to-date" in res.output


def test_create_merged_vcf__needs_two_samples(mocker):
    """Test create merged file failure when there's only one sample."""
    project_id = str(uuid4())
    mocked_response = ["should have at least 2 samples"]

    runner = CliRunner()

    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_create_merged_vcf = mocker.patch.object(
        APIClient,
        "create_merged_vcf",
        return_value=mocked_response,
        side_effect=APIClientError(
            message=mocked_response[0], status_code=400
        ),
    )

    res = runner.invoke(
        create_merged_vcf,
        [
            project_id,
            "--email",
            "foo@bar.com",
            "--password",
            "123",
        ],
    )

    assert res.exit_code == 1
    mocked_login.assert_called_once()
    mocked_create_merged_vcf.assert_called_once()
    assert "should have at least 2 samples" in res.output


def test_create_merged_vcf__no_impute_files(mocker):
    """Test create merged file failure when there are no impute files."""
    project_id = str(uuid4())
    mocked_response = ["must have impute VCF files associated"]

    runner = CliRunner()

    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_create_merged_vcf = mocker.patch.object(
        APIClient,
        "create_merged_vcf",
        return_value=mocked_response,
        side_effect=APIClientError(
            message=mocked_response[0], status_code=400
        ),
    )

    res = runner.invoke(
        create_merged_vcf,
        [
            project_id,
            "--email",
            "foo@bar.com",
            "--password",
            "123",
        ],
    )

    assert res.exit_code == 1
    mocked_login.assert_called_once()
    mocked_create_merged_vcf.assert_called_once()
    assert "must have impute VCF files associated" in res.output


def test_create_merged_vcf__success(mocker):
    """Test create merged file success."""
    project_id = str(uuid4())
    mocked_response = {
        "id": project_id,
        "created": "2020-09-14T08:59:00.480+00:00",
        "user": str(uuid4()),
        "last_status": {
            "id": str(uuid4()),
            "status": "running",
            "note": "",
            "created": "2020-07-28T12:46:22.719862+00:00",
        },
        "up_to_date": False,
    }

    runner = CliRunner()

    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_create_merged_vcf = mocker.patch.object(
        APIClient,
        "create_merged_vcf",
        return_value=ProjectMergeVCFs(**mocked_response),
    )

    res = runner.invoke(
        create_merged_vcf,
        [
            project_id,
            "--email",
            "foo@bar.com",
            "--password",
            "123",
        ],
    )

    assert res.exit_code == 0
    mocked_login.assert_called_once()
    mocked_create_merged_vcf.assert_called_once()
    output_line = io.BytesIO()
    sys.stdout = output_line
    echo(
        "\t".join(
            [
                mocked_response["id"],
                mocked_response["last_status"]["created"],
                mocked_response["last_status"]["status"],
            ]
        )
    )
    assert output_line.getvalue() in res.output.encode()
    assert (
        "Issued merge request for project {}".format(project_id) in res.output
    )



"""Base command layout.

All commands must implement this interface.
"""
import traceback

import click

from gencove.client import APIClient, APIClientError
from gencove.exceptions import ValidationError
from gencove.logger import (
    DEBUG,
    LOG_LEVEL,
    dump_debug_log,
    echo_data,
    echo_debug,
    echo_error,
    echo_info,
    echo_warning,
)
from gencove.utils import login, validate_credentials


class Command(object):  # pylint: disable=R0205
    """Base command structure."""

    def __init__(self, credentials, options):
        self.api_client = APIClient(options.host)
        self.is_logged_in = False
        self.credentials = credentials
        self.options = options
        self.is_credentials_valid = validate_credentials(credentials)

    def initialize(self):
        """Put any initializing logic here, such as login."""
        raise NotImplementedError

    def validate(self):
        """Validate that the arguments and initialize were successful."""
        raise NotImplementedError

    def execute(self):
        """Execute command logic."""
        raise NotImplementedError

    def validate_login_success(self):
        """Check if login succeeded."""
        if not self.is_logged_in:
            raise ValidationError(
                "Please check your credentials and try again"
            )

    def run(self):
        """Run the command.

        No need to override this, unless more customized behaviour is needed.
        """
        try:
            self.initialize()
            self.validate()
            self.validate_login_success()
            self.execute()
        except ValidationError as err:
            self.echo_error(err.message)
            dump_debug_log()
            raise click.Abort()
        except APIClientError as err:
            if err.status_code == 403:
                self.echo_error(
                    "You do not have the sufficient permission "
                    "level required to perform this operation."
                )
            self.echo_error(err.message)
            dump_debug_log()
            if LOG_LEVEL == DEBUG:
                raise err
            raise click.Abort()
        except Exception as err:
            self.echo_error(traceback.format_exc())
            dump_debug_log()
            if LOG_LEVEL == DEBUG:
                raise err
            raise click.Abort()

    def login(self):
        """Login current user."""
        if not self.is_credentials_valid:
            return False
        self.is_logged_in = login(self.api_client, self.credentials)
        return self.is_logged_in

    @staticmethod
    def echo_data(msg, **kwargs):
        """Output data message."""
        echo_data(msg, **kwargs)

    @staticmethod
    def echo_info(msg, **kwargs):
        """Output info message."""
        echo_info(msg, **kwargs)

    @staticmethod
    def echo_warning(msg, **kwargs):
        """Output warning message."""
        echo_warning(msg, **kwargs)

    @staticmethod
    def echo_error(msg, **kwargs):
        """Output error message."""
        echo_error(msg, **kwargs)

    @staticmethod
    def echo_debug(msg, **kwargs):
        """Output debug message."""
        echo_debug(msg, **kwargs)



"""Common utils used in multiple commands."""
import json
import uuid


def sanitize_string(output):
    """Removes unwanted characters from output string."""
    return output.replace("\t", " ")


def is_valid_uuid(candidate):
    """Test if provided string is a valid uuid version 4 string.

    candidate (str): uuid to check

    Returns:
        bool: True if is a valid uuid v4, False if not
    """
    try:
        uuid.UUID(candidate, version=4)
        return True
    except ValueError:
        return False


def is_valid_json(candidate):
    """Test if provided string is a valid JSON.

    candidate (str): JSON to check

    Returns:
        bool: True if is a valid JSON, False if not
    """
    try:
        json.loads(candidate)
        return True
    except ValueError:
        return False
